#version 150

//Made by DartCat25
//Thanks to people from ShaderLABS

#moj_import <three-dimensional-particles/config.glsl>
#moj_import <three-dimensional-particles/game_version.glsl>

#moj_import <fog.glsl>
#moj_import <light.glsl>

//Cubes
#define CUBE_SIZE 1 / 16.0
#define LIGHT_POWER 1.0

uniform sampler2D Sampler0;

uniform vec4 ColorModulator;
uniform float FogStart;
uniform float FogEnd;
uniform vec4 FogColor;
uniform mat4 ModelViewMat;
uniform mat4 ProjMat;

#if GAME_VERSION == 0
uniform vec3 Light0_Direction;
uniform vec3 Light1_Direction;
#else
vec3 Light0_Direction = vec3(0.0, 0.784, 0.549);
vec3 Light1_Direction = vec3(0.149, 0.745, 0.0);
#endif

in float vertexDistance;
in vec2 texCoord0;
in vec4 vertexColor;

in vec4 Pos;
in vec4 glPos;

in vec2 uv1, uv2, inUV;
flat in vec2 uv3;
in vec3 particle3DBoxSize;
flat in int ditheredOpacity;

out vec4 fragColor;

mat4 dither_matrix = mat4(
    // https://discourse.urho3d.io/t/screen-door-transparency/5054/2
    1.0 / 17.0,   9.0 / 17.0,   3.0 / 17.0,   11.0 / 17.0,
    13.0 / 17.0,  5.0 / 17.0,   15.0 / 17.0,  7.0 / 17.0,
    4.0 / 17.0,   12.0 / 17.0,  2.0 / 17.0,   10.0 / 17.0,
    16.0 / 17.0,  8.0 / 17.0,   14.0 / 17.0,  6.0 / 17.0
);

//Got from https://iquilezles.org/articles/intersectors/
vec3 sBox(vec3 ro, vec3 rd, vec3 size, out vec3 outNormal)
{
    vec3 m = 1.0 / rd;
    vec3 n = m * ro;
    vec3 k = abs(m) * size;
    vec3 t1 = -n - k;
    vec3 t2 = -n + k;
    float tN = max( max( t1.x, t1.y ), t1.z );
    float tF = min( min( t2.x, t2.y ), t2.z );
    if( tN > tF || tF < 0.0) discard;

    outNormal = -sign(rd)*step(t1.yzx,t1.xyz)*step(t1.zxy,t1.xyz);
    
    vec3 pos = (ro + rd * tN) / size;
    vec2 tex = vec2(0);
    if (abs(outNormal.x) > 0.9)
        tex = pos.zy;
    else if (abs(outNormal.y) > 0.9)
        tex = pos.xz;
    else if (abs(outNormal.z) > 0.9)
        tex = pos.xy;

    return vec3(tex / 4 + 0.5, tN);
}

void main() {
    vec4 texFor3D = vec4(0.);
    vec4 tex = texture(Sampler0, texCoord0);
    bool is3D = greaterThan(particle3DBoxSize, vec3(0.)) == bvec3(true);
    // 3d particle code
    if (is3D) {
        //Box Raytracing
        vec3 dir = normalize((-glPos).xyz);
        vec3 normal = vec3(0);
        vec3 tex = sBox(Pos.xyz, dir, particle3DBoxSize, normal);
        
        //Texture
        vec2 auv1 = round(uv1 / (uv3.x == 0 ? 1 - inUV.x : 1 - inUV.y)); //Right-up corner
        vec2 auv2 = round(uv2 / (uv3.x == 0 ? inUV.y : inUV.x)); //Left-down corner

        ivec2 res = ivec2(abs(auv1 - auv2)); //Resolution of texture
        ivec2 stp = ivec2(min(auv1, auv2)); //Left-Up corner

        vec2 texSize = textureSize(Sampler0, 0);

        texFor3D = texture(Sampler0, (stp + tex.xy * res) / texSize);
        vec4 color = texFor3D * vertexColor * ColorModulator;
        if (color.a < 0.1)
            discard;

        //Light
        color = minecraft_mix_light(Light0_Direction, Light1_Direction, -normal, color);

        fragColor = linear_fog(color, vertexDistance, FogStart, FogEnd, FogColor);

        //Depth
        vec4 clipPos = ProjMat * ModelViewMat * vec4(-dir * tex.z, 1);
        gl_FragDepth = clamp(clipPos.z / clipPos.w * 0.5 + 0.5, 0, 10);
        if(ditheredOpacity <= 0) {
            return;
        }
    }
    gl_FragDepth = gl_FragCoord.z;

    // dithered opacity
    if(ditheredOpacity >= 1) {
        float a = is3D ? texFor3D.a : tex.a;
        int x = int(gl_FragCoord.x);
        int y = int(gl_FragCoord.y);
        if (a < dither_matrix[x % 4][y % 4])
            discard;
        texFor3D.a = 100.;
        tex.a = 100.;
        if(is3D) {
            return;
        }
    }

    // vanilla code
    vec4 color = tex * vertexColor * ColorModulator;
    if (color.a < 0.1)
        discard;
    fragColor = linear_fog(color, vertexDistance, FogStart, FogEnd, FogColor);

}
