#version 150

//Made by DartCat25
//Thanks to people from ShaderLABS

#moj_import <three-dimensional-particles/util.glsl>
#moj_import <three-dimensional-particles/config.glsl>
#define UV_OFFSET 1./12000.

#moj_import <fog.glsl>

in vec3 Position;
in vec2 UV0;
in vec4 Color;
in ivec2 UV2;

uniform sampler2D Sampler0;
uniform sampler2D Sampler2;

uniform mat4 ModelViewMat;
uniform mat4 ProjMat;
uniform int FogShape;

out float vertexDistance;
out vec2 texCoord0;
out vec4 vertexColor;

out vec4 Pos;
out vec4 glPos;

out vec2 uv1, uv2, inUV;
flat out vec2 uv3;
out vec3 particle3DBoxSize;
flat out int ditheredOpacity;

#define CALC_GL_POS gl_Position = ProjMat * ModelViewMat * vec4(Position, 1.0);

void main() {
    vertexDistance = fog_distance(Position, FogShape);
    texCoord0 = UV0;
    vertexColor = Color * texelFetch(Sampler2, UV2 / 16, 0);

    Pos = vec4(0);
    vec2 texSize = textureSize(Sampler0, 0);
    vec2 uv = (UV0 * texSize);
    int id = gl_VertexID % 4;
    
    const vec2[4] corners = vec2[4](vec2(0), vec2(0, 1), vec2(1, 1), vec2(1, 0));
    const vec2[4] uvOffset = vec2[4](vec2(UV_OFFSET, UV_OFFSET), vec2(UV_OFFSET, -UV_OFFSET), vec2(-UV_OFFSET, -UV_OFFSET), vec2(-UV_OFFSET, UV_OFFSET));
    vec4 markers = texture(Sampler0, UV0 - uvOffset[id]);

    particle3DBoxSize = vec3(-1.);
    vec3 particlePosOffset = vec3(0.);
    float quadExpandMultiplier = 1.;
    float tracedShapeExpandMultiplier = 1.6;
    vec2 uv1Offset = vec2(0.);
    vec2 uv2Offset = vec2(0.);
    float lightOffsetColorMultiplier = 1.5; // multiplied by Color to account for pre multiplied lighting on some particles
    ditheredOpacity = 0;

    bool is3DParticle = rougheq(markers.rga, vec3(28./255., 21./255., 3./255.));
    bool isBlockBreakingParticle = floor(uv) != uv && texSize.y / texSize.x != 4;

    // return early if not 3d
    if((!is3DParticle && !isBlockBreakingParticle) || vertexDistance > PARTICLE_3D_DISTANCE) {
        CALC_GL_POS
        return;
    }

    if (vertexDistance < PARTICLE_3D_DISTANCE && is3DParticle && rougheq(markers.b, 1./255.)) {
        // drip particle
        // (8x8 texture with 2x2 pixels in the middle)
        particle3DBoxSize = vec3(0.75/24.);
        particlePosOffset = vec3(0., 0.5/16., 0.);

        uv1Offset = vec2(-3.);
        uv2Offset = vec2(3.);
    }
    else if (vertexDistance < PARTICLE_3D_DISTANCE && is3DParticle && rougheq(markers.b, 2./255.)) {
        // single pixel particle (used for things such as crimson_spore)
        // (8x8 texture with 1 pixel in the middle closer to bottom right corner)
        particle3DBoxSize = vec3(0.4/24.);
        particlePosOffset = vec3(0., 1./16., 0.);

        uv1Offset = vec2(-3.);
        uv2Offset = vec2(4.);
    }
    else if (vertexDistance < PARTICLE_3D_DISTANCE && is3DParticle && inrange(markers.b, 10./255., 20./255.)) {
        // size controlled cube (blue channel between 10 and 20)
        // (8x8 texture with 2x2 pixels in the middle)
        particle3DBoxSize = vec3((markers.b)/2.);

        uv1Offset = vec2(-3.);
        uv2Offset = vec2(3.);
    }
    else if (vertexDistance < PARTICLE_3D_DISTANCE && is3DParticle && inrange(markers.b, 21./255., 29./255.)) {
        // height controlled cuboid (blue channel between 21 and 29)
        // this is specifically for trial_spawner_detection and will probably not work well with other things
        float cuboidHeight = ((markers.b*255.)-20.);
        particle3DBoxSize = vec3(1./48., cuboidHeight/48., 1./48.);
        particlePosOffset = vec3(0, 0.0625, 0);
        quadExpandMultiplier = 2.5;
        tracedShapeExpandMultiplier = 2.5;

        switch (int(floor(cuboidHeight))) {
            case 6:
                uv1Offset = vec2(-2 * 2., 2.); // 6 tall
                uv2Offset = vec2(1.5 * 2., -2.);
                break;
            case 4:
                uv1Offset = vec2(-2 * 2., -1.); // 4 tall
                uv2Offset = vec2(1.5 * 2., -1.);
                break;
            case 3:
                uv1Offset = vec2(-2 * 2., -3.); // 3 tall
                uv2Offset = vec2(1.5 * 2., 0.);
                break;
            case 1:
                uv1Offset = vec2(-2 * 2., -7.); // 1 tall
                uv2Offset = vec2(1.5 * 2., 1.);
                break;
        }
    }
    else if (vertexDistance < PARTICLE_3D_DISTANCE && is3DParticle && inrange(markers.b, 30./255., 40./255.)) {
        // same as size controlled cube but with dithered opacity (blue channel between 30 and 40)
        // (8x8 texture with 2x2 pixels in the middle)
        particle3DBoxSize = vec3((markers.b - (20./255.))/2.);
        ditheredOpacity = 1;

        uv1Offset = vec2(-3.);
        uv2Offset = vec2(3.);
    }

    if (vertexDistance < PARTICLE_3D_DISTANCE && isBlockBreakingParticle) {
        // block breaking particles
        particle3DBoxSize = vec3(1./16.);
        particlePosOffset = vec3(0, 0.0625, 0);
        lightOffsetColorMultiplier = 1.75;
    }

    if(vertexDistance < PARTICLE_3D_DISTANCE && (isBlockBreakingParticle || is3DParticle)) {
        // general 3d particle logic
        vec2 inCoords = (corners[id] - 0.5) / 4 * vec2(1, -1);
        
        Pos = vec4(Position + particlePosOffset, 1);
        glPos = Pos - vec4(inCoords * tracedShapeExpandMultiplier, 0, 0) * ModelViewMat;
        gl_Position = ProjMat * (ModelViewMat * Pos - vec4(inCoords * quadExpandMultiplier, 0, 0)); //Expand Particle

        uv1 = uv2 = vec2(0);
        if (id == 0) uv1 = uv + uv1Offset;
        if (id == 2) uv2 = uv + uv2Offset;
        uv3 = inUV = corners[id];
        
        vertexColor = Color * vec4(vec3(lightOffsetColorMultiplier), 1.) * texelFetch(Sampler2, UV2 / 16, 0);
        return;
    }
    CALC_GL_POS
}
